#include"writer.h"

#include<string.h>
#include<endian.h>
#include<stdlib.h>

static uint64_t VARINT_MASKS[] = {0, 0x80L, 0xC000L, 0xE00000L, 0xF0000000L, 0xF800000000L, 0xFC0000000000L, 0xFE000000000000L, 0L};

static int byte_length(uint64_t i) {
	if(i == 0) return 1;
	
	int n = 0;
	while(i) {
		i >>= 8;
		n++;
	}
	return n;
}

static int bit_length(uint64_t i) {
	if(i == 0) return 1;
	
	int n = 0;
	while(i) {
		i >>= 1;
		n++;
	}
	
	return n;
}

static uint64_t encode_int(uint64_t i) {
	return i | VARINT_MASKS[(bit_length(i) + 6) / 7];
}

static void advance(EBMLWriter *this, uint64_t amount) {
	for(int i = 0; i < this->currentDepth; i++) {
		this->stack[i] += amount;
	}
}

static void add(EBMLWriter *this, const void *data, size_t length) {
	if(this->bufferLen + length > this->bufferCapacity) {
		this->buffer = realloc(this->buffer, (this->bufferLen + length + 1023) & ~1023);
	}
	
	memcpy(this->buffer + this->bufferLen, data, length);
	
	this->bufferLen += length;
	
	advance(this, length);
}

static void add_varint(EBMLWriter *this, uint64_t i) {
	int len = byte_length(i);
	
	i = htobe64(i);
	
	add(this, (uint8_t*) &i + 8 - len, len);
}

void ebml_writer_init(EBMLWriter *this) {
	memset(this, 0, sizeof(EBMLWriter));
	
	this->buffer = calloc(this->bufferCapacity = 1024, 1);
}

void ebml_writer_push(EBMLWriter *this, uint64_t id) {
	add_varint(this, id);
	
	add(this, &(uint64_t) {htobe64(0x0100000000000000L)}, sizeof(uint64_t)); // LENGTH OF MAX SIZE (BECAUSE UNKNOWN)
	
	this->stack[this->currentDepth++] = 0;
}

void ebml_writer_pop(EBMLWriter *this) {
}

void ebml_writer_put(EBMLWriter *this, uint64_t id, EBMLElementType type, EBMLPrimitive primitive) {
	add_varint(this, id);
	
	int len = byte_length(id);
	
	if(type == EBML_UNSIGNED_INTEGER) {
		len = byte_length(primitive.uInt);
	} else if(type == EBML_FLOAT4) {
		len = 4;
	} else if(type == EBML_FLOAT8) {
		len = 8;
	} else if(type == EBML_STRING) {
		len = strlen(primitive.string);
	} else if(type == EBML_DATE) {
		len = 8;
	} else abort();
	
	add_varint(this, encode_int(len));
	
	if(type == EBML_UNSIGNED_INTEGER) {
		primitive.uInt = htobe64(primitive.uInt);
		add(this, (uint8_t*) &primitive.uInt + 8 - len, len);
	} else if(type == EBML_FLOAT4) {
		uint32_t fl = *(uint32_t*) &primitive.flt4;
		add(this, &fl, 4);
	} else if(type == EBML_FLOAT8) {
		uint64_t fl = *(uint64_t*) &primitive.flt8;
		add(this, &fl, 8);
	} else if(type == EBML_STRING) {
		add(this, primitive.string, len);
	} else if(type == EBML_DATE) {
		add(this, &primitive.date, 8);
	} else abort();
}
